%{
#include <stdio.h>
#include <string.h> 
#include "ipp.h"

int act_err;
char *act_err_par;
fact_list act_var, act_const;
effect_list q_goal_list;
token_list act_goal;
fact_list act_pardec;
fact_list new_neg_pred;

#ifndef SCAN_ERR
#define SCAN_ERR
#define TYPELIST_EXPECTED            0
#define INITIALTOK_EXPECTED          1
#define FACTLIST_EXPECTED            2
#define SEMICOLON_EXPECTED           3
#define GOALTOK_EXPECTED             4
#define	GOALS_EXPECTED               5
#define TYPENAME_EXPECTED            6
#define CONSTLIST_EXPECTED           7
#define PARDEC_EXPECTED              8
#define GOALCOND_OR_IMPL_EXPECTED    9
#define MISSING_GOALDES             10
#define TYPENAME_UNKNOWN            11
#endif

static char *errmsg[] = {
  "typelist expected",
  "\'initial:\' or typelist expected",
  "list of facts expected",
  "\';\' expected",
  "\'goal:\' expected",
  "list of goals expected",
  "typename ( \'type\' : ) expexted",
  "list of constants expected",
  "parameter declaration for quantifier expected",
  "goal conditions or \'=>' expected",
  "missing goal description",
  "unknown type '%s' (check for typing errors...)",
  NULL
};

void fcterr( int errno, char *par );
BOOLEAN cond_in_initial( token_list conds, token_list initials );
effect_list make_q_goal_list( fact_list pars, fact_list conds, 
			      fact_list goals, int quant );

%}



%start factfile

%union {
  char string[256];
  fact_list fact_list;
  op_list op_list;
  token_list token_list;
  effect_list effect_list;
}

%type <fact_list> typelist
%type <effect_list> goals
%type <fact_list> constantlist
%type <fact_list> factlist
%type <fact_list> pardec
%type <fact_list> varlist
%type <token_list> args



%token <string> VARNAME 
%token <string> COMMENT
%token <string> CONSTNAME
%token <string> PREDNAME
%token <string> TYPENAME
%token <string> QUANT
%token IMPL
%token NOT
%token INI
%token GOA


%%

/* numbers in comments are only useful for the programmer
   and reader of the yacc source code. Each terminal or
   non-terminal symbol and (which is not nice) each
   program section in {brackets} has a number and may
   be reached by the $ operator - the numbers only
   help prevent counting all the time and produce
   ugly fat spots on the monitior */


factfile        : { fcterr( TYPELIST_EXPECTED, NULL ); } typelist /* 1 2 */
                  { orig_constant_list = $2;
		    /* build a fact list containing token_lists for all types */
		    build_type_list();		    
		    fcterr( INITIALTOK_EXPECTED, NULL ); } INI /* 3 4 */
                  { fcterr( FACTLIST_EXPECTED, NULL ); } factlist 
                  { fcterr( SEMICOLON_EXPECTED, NULL ); } ';' 
                  { orig_initial_facts = $6;
		    fcterr( GOALTOK_EXPECTED, NULL ); } GOA /* 9 10 */
		  { fcterr( GOALS_EXPECTED, NULL ); } goals 
                  {
		    raw_goal_facts = $12;
		  } /* 13 */
		;


goals           : /* empty */
                    { $$ = NULL; } 
                | QUANT { fcterr( PARDEC_EXPECTED, NULL ); } pardec  /* 1 2 3 */
		  { fcterr( GOALCOND_OR_IMPL_EXPECTED, NULL ); } factlist IMPL /* 4 5 6 */
		  { fcterr( MISSING_GOALDES, NULL ); } factlist ';' goals 
                    {
		      if ( !$8 ) 
			{
			  fcterr( MISSING_GOALDES, NULL );
			  yyerror();
			}
		      if ( $1[0] == 'A' )
			$$ = make_q_goal_list( $3, $5, $8, ALL_QUANT );
		      else
			$$ = make_q_goal_list( $3, $5, $8, EX_QUANT );
		      $$->next = $10;
		    }
                | factlist IMPL factlist ';' goals 
                    {
		      $$ = make_q_goal_list( NULL, $1, $3, NO_QUANT );
		      $$->next = $5;
		    }
                | factlist ';' goals 
                    { 
		      $$ = make_q_goal_list( NULL, NULL, $1, NO_QUANT );
		      $$->next = $3;
		    } 
                ;

typelist        : /* empty */
                    { $$ = NULL; } 
                | { fcterr( TYPENAME_EXPECTED, NULL ); } TYPENAME ':'      /* 1 2 3 */ 
		  { fcterr( CONSTLIST_EXPECTED, NULL ); } constantlist ';' /* 4 5 6 */
		  { fcterr( TYPELIST_EXPECTED, NULL ); } typelist          /* 7 8 */
                    {
		      $$ = $5;
		      for( act_const=$$; act_const; act_const=act_const->next )
			{ /* adjust types of constant list */
			  act_const->item->next = 
			    ( token_list ) malloc( sizeof( token_list_elt ) );
			  act_const->item->next->item = 
			    ( token ) calloc( 1 + strlen( $2 ), sizeof( char ) );
			  strcpy( act_const->item->next->item, $2 );
			  act_const->item->next->next = NULL;
			  if ( !act_const->next ) {
			    act_const->next = $8; 
			    break;
			  }
			}
		    }
                ;

constantlist    : CONSTNAME 
                    {
		      $$ = ( fact_list ) malloc( sizeof( fact_list_elt ) );
		      $$->item = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item->item = 
			( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( $$->item->item, $1 );
		      $$->next = NULL;
		    }
                | CONSTNAME constantlist 
                    {
		      $$ = ( fact_list ) malloc( sizeof( fact_list_elt ) );
		      $$->item = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item->item = 
			( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( $$->item->item, $1 );
		      $$->next = $2;
		    }
                ;


pardec          : varlist CONSTNAME 
                    {
		      $$ = $1;
		      act_pardec = $$;
		      if ( !type_known( $2 ) )
			{
			  fcterr( TYPENAME_UNKNOWN, $2 );
			  yyerror();
			}
		      for( act_var=$$; act_var; act_var=act_var->next )
			{ /* adjust types (2nd element of token_list) for varlist */
			  act_var->item->next = 
			    ( token_list ) malloc( sizeof( token_list_elt ) );
			  act_var->item->next->item = 
			    ( token ) calloc( 1 + strlen( $2 ), sizeof( char ) );
			  strcpy( act_var->item->next->item, $2 );
			  act_var->item->next->next = NULL;
			}
		    }
                | varlist CONSTNAME 
                  {
		    if ( !type_known( $2 ) )
		      {
			fcterr( TYPENAME_UNKNOWN, $2 );
			yyerror();
		      }
		  }
		  pardec
                    {
		      $$ = $1;
		      act_pardec = $$;
		      for( act_var=$$; act_var; act_var=act_var->next )
			{ /* adjust types (2nd element of token_list) for varlist */
			  act_var->item->next = 
			    ( token_list ) malloc( sizeof( token_list_elt ) );
			  act_var->item->next->item = 
			    ( token ) calloc( 1 + strlen( $2 ), sizeof( char ) );
			  strcpy( act_var->item->next->item, $2 );
			  act_var->item->next->next = NULL;
			  if ( !act_var->next ) {
			    act_var->next = $4; 
			    break;
			  }
			}
		    }
		;
	  
varlist         : VARNAME
                    {
		      $$ = ( fact_list ) malloc( sizeof( fact_list_elt ) );
		      $$->item = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item->item = 
			( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( $$->item->item, $1 );
		      $$->next = NULL;
		    }
	        | VARNAME varlist
                    {
		      $$ = ( fact_list ) malloc( sizeof( fact_list_elt ) );
		      $$->item = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item->item = 
			( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( $$->item->item, $1 );
		      $$->next = $2;
		    }
		;
      
factlist       : /* empty */ 
                    { $$ = NULL; }
                | PREDNAME '(' args ')' factlist
                    {
		      $$ = ( fact_list ) malloc( sizeof( fact_list_elt ) );
		      $$->item = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item->item = 
			( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( $$->item->item, $1 );
		      $$->item->next = $3;
		      $$->next = $5;
             	    }
                | NOT PREDNAME '(' args ')' factlist
                    {
		      $$ = ( fact_list ) malloc( sizeof( fact_list_elt ) );
		      $$->item = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item->item = 
			( token ) calloc( 2 + strlen( $2 ), sizeof( char ) );
		      strcpy( $$->item->item, "!" );
		      strcat( $$->item->item, $2 );
		      $$->item->next = $4;
		      $$->next = $6;
                    		             	    }
                ; 

args            : /* empty */
                    { $$ = NULL; }
                | CONSTNAME args
                    {
		      $$ = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item = ( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( $$->item, $1 );
		      $$->next = $2;
		    }
                | VARNAME args
                    {
		      token_list t;
		      t = ( token_list ) malloc( sizeof( token_list_elt ) );
		      t->item = ( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( t->item, $1 );
		      t->next = $2;
		      $$ = t;
		    }
		;      


%%

/* lexer must be built with:  flex -Pfct lex-fct.l */
#include "lex.fct.c"

void fcterr( int errno, char *par )
{
  act_err = errno;
  act_err_par = par;
}
  
int yyerror( char *msg )
{
  printf("%s: syntax error in line %d, '%s':\n", act_filename, lineno, yytext );
  if ( act_err_par )
    {
      printf( errmsg[act_err], act_err_par );
      printf( "\n" );
    }
  else
    printf("%s\n", errmsg[act_err] );
  exit( 1 );
}

  
void load_fct_file( char *filename )
{
  FILE *fp;/* pointer to input files */

  /* open fact file */
  if( ( fp = fopen( filename, "r" ) ) == NULL ) {
    printf( "\nipp: can't find fact file: %s\n\n", filename );
    exit( 1 );
  }
  act_filename = filename;
  lineno = 1; 
  yyin = fp;
  yyparse();
  fclose( fp );/* and close file again */
}


effect_list make_q_goal_list( fact_list pars, fact_list conds, 
			      fact_list goals, int quant )
{
  effect_list result;
  
  result = ( effect_list ) calloc( 1, sizeof( effect_list_elt ) );
  result->quantified_variables = ( fact_list ) calloc( 1, sizeof( fact_list_elt ) );
  result->quantified_variables->item = 
    ( token_list ) calloc( 1, sizeof( token_list_elt ) );
  result->quantified_variables->item->item = ( token ) calloc( 1, sizeof( char ) );
  result->quantified_variables->item->item[0] = QUANTIFIERS[quant];
  result->quantified_variables->next = pars;
  result->conditions = conds;
  result->add_effects = goals;
  result->del_effects = NULL;
  result->next = NULL;
  return result;
}


void print_factlist( fact_list list )
{

  fact_list i_list;
  token_list i_token;

  for ( i_list = list; i_list; i_list = i_list->next ) 
    for ( i_token = i_list->item; i_token; i_token = i_token->next )
      printf( "%s ", i_token->item );
}

/* prints out facts */
void print_fct( fact_list con_list, token_list ini_list, token_list gol_list )

{

  printf( "\nconstants: " );
  print_factlist( con_list ); 

  printf( "\n\ninitial: " );
  print_tokenlist( ini_list );

  printf( "\n\ngoal: " );
  print_tokenlist( gol_list );

  printf( "\n" );

}

  
void print_tokenlist(token_list list)
{
  token_list i_token;

  for ( i_token = list; i_token; i_token = i_token->next )
    {
      printf( "\n%s ", i_token->item );
      if (!strcmp(i_token->item, ""))
	printf("Match\n");
    }
}
