/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 */

/* 	$Id: utilities.c,v 1.4 1997/10/29 16:34:38 ipp Exp $	 */

#ifndef lint
static char vcid[] = "$Id: utilities.c,v 1.4 1997/10/29 16:34:38 ipp Exp $";
#endif /* lint */


#include"ipp.h"

/* These are helper functions that might be used everywhere */

fact_list new_fact_list(void)
{
  fact_list result;

  result = ( fact_list ) malloc( sizeof( fact_list_elt ) );
  if (!result)
    {
      fprintf(stderr, "%s", NO_MEMORY);
      exit(2);
    }

  result->item = NULL; 
  result->next = NULL;
  return result;
}

token_list new_token_list(void)
{
  token_list result;
  result = ( token_list ) malloc( sizeof( token_list_elt ) );
  if (!result)
    {
      fprintf(stderr, "%s", NO_MEMORY); 
      exit(2);
    }

  result->item = NULL; 
  result->next = NULL;
  return result;
}

effect_list new_effect_list(void)
{
  effect_list result;

  result = ( effect_list ) malloc( sizeof( effect_list_elt ) );
  if (!result)
    {
      fprintf(stderr, "%s", NO_MEMORY);
      exit(2);
    }

  result->quantified_variables = NULL;
  result->conditions = NULL;
  result->del_effects = NULL;
  result->add_effects = NULL;
  return result;
}

token new_token(int size)
{
  char* tok;

  tok = (char*) calloc(size, sizeof(char));
  if (!tok)
    {
      fprintf(stderr, "%s", NO_MEMORY);
      exit(2);
    }
  return tok;
}

operator_list new_ground_operator( token name )
{
  operator_list act_op;
  act_op = ( operator_list ) malloc( sizeof( operator_list_elt ) );
  act_op->name = new_token( strlen( name ) + 1 );
  strcpy( act_op->name, name );
  act_op->preconditions = NULL;
  act_op->effects = NULL;
  act_op->objects = NULL;
  act_op->next = NULL;
  return act_op;
}

/* 'blowing up' a token list to get a fact list in which all
   token_lists only contain one token ... 
   this is only needed to be able to pass operator instatiation
   even when a (ex-goal) operator in 'uninstatiated format doesn't 
   need instatiation at all */
void factlist_from_tokenlist( fact_list *fl, token_list tl )
{
  fact_list *act_f;
  token_list act_t;

  for( act_f=fl, act_t=tl; act_t; 
       act_f = &((*act_f)->next), act_t=act_t->next )
    {
      *act_f = new_fact_list();
      (*act_f)->item = new_token_list();
      (*act_f)->item->item = act_t->item;
    }
}

/*
 * This function copies a token_list by creating a new
 * token_list_elt for every element in source and
 * copying the item and next pointer to the new token_list_elt.
 * Then the function calls itself with the next element of source.
 *
 * returns: the start of the new token_list
 */
token_list copy_token_list( token_list source )
{
  token_list temp;

  if ( !source )
    {
      temp = NULL;
    }
  else
    {
      temp = new_token_list();
      temp->item = new_token( strlen( source->item ) + 1 );
      
      strcpy( temp->item, source->item );
      
      temp->next = copy_token_list( source->next );
    }
  return temp;
}


operator_list copy_operator_list( operator_list source )
{ 
  operator_list temp;

  if ( !source )
    {
      temp = NULL;
    }
  else
    {
      temp = new_ground_operator( source->name );
      temp->preconditions = source->preconditions;
      temp->effects = source->effects;
      temp->objects = source->objects;
      temp->next = copy_operator_list( source->next );
    }
  return temp;
}
